// kindlescreen.cpp
//
// Copyright (C) 2010, 2009 Griffin I'Net, Inc. (blog@griffin.net)
//
// This file is licensed under the LGPL version 2.1, the text of which should
// be in the LICENSE.txt file, or alternately at this location:
// http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
//
// DISCLAIMER: This software is released AS-IS with ABSOLUTELY NO WARRANTY OF
// ANY KIND.  The use of this software indicates your knowledge and acceptance
// of this.

#include <fcntl.h>
#include <asm/types.h>
typedef __u8 u8;
#include <linux/einkfb.h>
#include <sys/ioctl.h>
#include "kindlescreen.h"
#include <QFile>

#include "kindlescreenthread.h"


KindleScreen::KindleScreen(int display_id, bool direct) : QLinuxFbScreen(display_id)
{
    direct = !direct;

    if (direct)
        _type = 0;
    else
        _type = 1;

    _dirty = false;

    _dataLast = NULL;
}

KindleScreen::~KindleScreen()
{
    // NOP
}

bool KindleScreen::initDevice()
{
    qDebug("KindleScreen::initDevice() %s", (flashingupdates == true)? "flashing updates": "partial updates");


    _fbd = open(EINK_FRAME_BUFFER, O_RDWR);

    if (_fbd == -1)
        qDebug("FAILED to open framebuffer.");

    if (_type == 0)
    {
        _thread = new KindleScreenThread(this);
        _thread->start();
    }
    else
        _thread = NULL;

    if (!QLinuxFbScreen::initDevice())
        return false;

//    QRgb* table = clut();
//
//    qDebug("Color table count: %d", colorCount());
//
//    for(int i=0;i<colorCount();i++)
//        qDebug("%d : %08X", i, (unsigned int)table[i]);

    if ((data != NULL) && (size > 0))
    {
        _dataLast = (uchar*)malloc(size);
        memcpy(_dataLast, data, size);
    }

    myCursor = new KindleCursor() ;
    myCursor->setScreen_fd(_fbd) ;
    qt_screencursor =  myCursor ;

    return true;
}


bool KindleScreen::connect ( const QString & displaySpec )
{
    if (displaySpec.contains("flashing", Qt::CaseInsensitive))
    {
        flashingupdates = true ;
    }
    else
        flashingupdates = false ;

    return QLinuxFbScreen::connect(displaySpec)  ;
}


int KindleScreen::alloc(unsigned int red, unsigned int green, unsigned int blue)
{
//    return colorCount() - 1 - QLinuxFbScreen::alloc(red, green, blue);
    return QLinuxFbScreen::alloc(red, green, blue);
}

void KindleScreen::shutdownDevice()
{
//    qDebug("KindleScreen::shutdownDevice()");

    // TODO : stop screen update service here
    _thread->quitNow();
    delete _thread;

    QLinuxFbScreen::shutdownDevice();

    close(_fbd);

    if (_dataLast != NULL)
    {
        free(_dataLast);
        _dataLast = NULL;
    }
}

void KindleScreen::solidFill(const QColor& color, const QRegion& region)
{
    QLinuxFbScreen::solidFill(color, region);
}

void KindleScreen::blit(const QImage& image, const QPoint& topLeft, const QRegion& region)
{
    QImage imageInverted = image;
    imageInverted.invertPixels();

    qDebug(">>>> blit image %dx%dx%d type=%d", image.width(),image.height(), image.depth() ,image.format()) ;

    QLinuxFbScreen::blit(imageInverted, topLeft, region);
}

/*
void KindleScreen::solidFill(const QColor& color, const QRegion& region)
{
    QLinuxFbScreen::solidFill(color, region);

    QRect r = region.boundingRect();

    if (!r.isNull())
    {
        qDebug("fill region: %d, %d, %d, %d", r.topLeft().x(), r.topLeft().y(), r.width(), r.height());
        updateRect(r);
    }
}

void KindleScreen::blit(const QImage& image, const QPoint& topLeft, const QRegion& region)
{
    QLinuxFbScreen::blit(image, topLeft, region);

    QRect b = region.boundingRect();

    if (!b.isNull())
    {
        QRect r(topLeft, QPoint(b.width() - 1, b.height() - 1));
        qDebug("region: %d, %d, %d, %d", r.topLeft().x(), r.topLeft().y(), r.width(), r.height());

        updateRect(r);
    }
}
*/

void KindleScreen::setDirty ( const QRect & r )
{
    //qDebug("setDirty: %d, %d, %d, %d", r.topLeft().x(), r.topLeft().y(), r.width(), r.height());
    updateRect(r);

    QLinuxFbScreen::setDirty(r);
    //qDebug("_____________");
}

void KindleScreen::updateRect ( const QRect & rc )
{
    QRect r = rc;

    qDebug("updateRect: %d, %d, %d, %d", r.topLeft().x(), r.topLeft().y(), r.width(), r.height());


    if (_type == 1)
    {
        _mutex.lock();
        _dirty = true;
        _updateRect = _updateRect.united(r);
        _mutex.unlock();
    }
    else
    {

        update_area_t ua;
        ua.x1 = r.left();
        ua.y1 = r.top();
        ua.x2 = r.right() + 1;
        ua.y2 = r.bottom() + 1;
        ua.which_fx = (flashingupdates == false) ? fx_update_partial : fx_update_full; //fx_invert
        ua.buffer = NULL;
        ioctl(_fbd, FBIO_EINK_UPDATE_DISPLAY_AREA, &ua);

    }
}

void KindleCursor::hide ()
{
    if (visible)
    {
        update_area_t ua;
        ua.x1 = curx ;
        ua.y1 = cury;
        ua.x2 = curx + curw ;
        ua.y2 = cury + curh ;
        ua.which_fx = fx_invert;
        ua.buffer = NULL;
        ioctl(screen_fbd, FBIO_EINK_UPDATE_DISPLAY_AREA, &ua);
        visible = false;
    }
}

void KindleCursor::move ( int x, int y )
{
    if (visible)
    {
        hide() ;
        curx = x ;
        cury = y ;
        show() ;
    }
    else
    {
        curx = x ;
        cury = y ;
    }
}

void KindleCursor::set ( const QImage & image, int hotx, int hoty )
{
    QScreenCursor::set(image, hotx, hoty) ;
    if (visible)
    {
        hide() ;
        curw = image.width() ;
        curx = image.height() ;
        show() ;
    }
    else
    {
        curw = image.width() ;
        curx = image.height() ;
    }
}

void KindleCursor::setBounds(int x0, int y0, int w, int h)
{
    if (visible)
    {
        hide() ;
        curx = x0 ;
        cury = y0 ;
        curw = w ;
        curx = h ;
        show() ;
    }
    else
    {
        curx = x0 ;
        cury = y0 ;
        curw = w ;
        curx = h ;
    }
}

void KindleCursor::show ()
{
    if (!visible)
    {
        update_area_t ua;
        ua.x1 = curx ;
        ua.y1 = cury;
        ua.x2 = curx + curw ;
        ua.y2 = cury + curh ;
        ua.which_fx = fx_invert;
        ua.buffer = NULL;
        ioctl(screen_fbd, FBIO_EINK_UPDATE_DISPLAY_AREA, &ua);
        visible = true;
    }
}

/***
QRect boundingRect () const
{
    return this->QScreenCursor::boundingRect() ;
}

QImage 	image () const
{
    return this->QScreenCursor::image() ;
}
***/
