// kindlescreen.h
//
// Copyright (C) 2010, 2009 Griffin I'Net, Inc. (blog@griffin.net)
//
// This file is licensed under the LGPL version 2.1, the text of which should
// be in the LICENSE.txt file, or alternately at this location:
// http://www.gnu.org/licenses/old-licenses/lgpl-2.1.txt
//
// DISCLAIMER: This software is released AS-IS with ABSOLUTELY NO WARRANTY OF
// ANY KIND.  The use of this software indicates your knowledge and acceptance
// of this.

#ifndef KINDLESCREEN_H
#define KINDLESCREEN_H


#include <qscreenlinuxfb_qws.h>
#include <QThread>
#include <QRect>
#include <QMutex>
#include <QRegion>
#include <QScreenCursor>

class KindleScreenThread;

class KindleCursor : public QScreenCursor
{
public:
    KindleCursor(int x=0, int y=0, int w=0, int h=0, bool visible=false) ;

    virtual ~KindleCursor() ;
    virtual void 	hide () ;
    virtual void 	move ( int x, int y ) ;
    virtual void 	set ( const QImage & image, int hotx, int hoty ) ;
    virtual void 	show () ;

    bool 	isAccelerated () const { return true; }
    bool 	isVisible () const { return visible; }
    QRect boundingRect () const ;
    QImage image () const ;
    void setScreen_fd(int fd) { screen_fbd = fd ; }

    void setBounds(int x0, int y0, int w, int h) ;

private:
    int curx ;
    int cury ;
    int curw ;
    int curh ;
    bool visible ;
    int screen_fbd ;
};

class KindleScreen : public QLinuxFbScreen
{
    friend class KindleScreenThread;
public:
    KindleScreen(int display_id, bool direct);
    virtual ~KindleScreen();

    virtual bool initDevice();
    virtual void shutdownDevice();

    virtual void solidFill(const QColor& color, const QRegion& region);
    virtual void blit(const QImage& image, const QPoint& topLeft, const QRegion& region);

    virtual void setDirty ( const QRect & rectangle );
    virtual void updateRect ( const QRect & rectangle );

    virtual int alloc(unsigned int red, unsigned int green, unsigned int blue);
    virtual bool connect(const QString &displaySpec);

protected:
    uchar* _dataLast;
    bool _dirty;
    QMutex _mutex;
    QRect _updateRect;
    int _fbd;
    int _type;
    KindleCursor *myCursor ;

private:
    KindleScreenThread* _thread;
    bool flashingupdates ;
};


#endif // KINDLESCREEN_H
